/*
 [The "BSD licence"]
 Copyright (c) 2013 Terence Parr, Sam Harwell
 Copyright (c) 2017 Ivan Kochurkin (upgrade to Java 8)
 All rights reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.
 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/**
 * Inspired by the Kotlin Lexer :
 *      https://raw.githubusercontent.com/antlr/grammars-v4/master/kotlin/kotlin/KotlinLexer.g4
 */

lexer grammar FeatherLexer;

import UnicodeClasses;

HashBangLine: '#!' .*? NL;

Documentation
    : '/**' .*? '*/'
    ;

DelimitedComment
    : '/*' ( DelimitedComment | . )*? '*/'
      -> channel(1)
    ;

LineComment
    : '//' ~[\u000A\u000D]*
      -> channel(1)
    ;

WS
    : [\u0020\u0009\u000C]
      -> skip
    ;

NL: '\u000A' | '\u000D' '\u000A' ;

//SEPARATORS & OPERATIONS

DOT: '.' ;
COMMA: ',' ;
LPAREN: '(' -> pushMode(Inside) ;
RPAREN: ')' ;
LSQUARE: '[' -> pushMode(Inside) ;
RSQUARE: ']' ;
BARRELOPEN: '[[' -> pushMode(Inside) ;
BARRELCLOSE: ']]' ;
LCURL: '{' ;
RCURL: '}' ;
POWER: '^';
MULT: '*' ;
REM: '%' ;
DIVIDE: '/' ;
BACKSLASH: '\\' ;
FLOOR_DIV: '~/' ;
ARROW: '->' ;
ADD: '+' ;
SUB: '-' ;
INCR: '++' ;
DECR: '--' ;
CONJ: '&&' ;
PIPE: '|' ;
DISJ: '||' ;
EXCL: '!' ;
COLON: ':' ;
COLON_COLON: '::' ;
CURRIED: ':>' ;
SEMICOLON: ';' ;
ASSIGNMENT: '=' ;
ADD_ASSIGNMENT: '+=' ;
SUB_ASSIGNMENT: '-=' ;
MULT_ASSIGNMENT: '*=' ;
FLOOR_DIV_ASSIGNMENT: '~/=' ;
DIVIDE_ASSIGNMENT: '/=' ;
REM_ASSIGNMENT: '%=' ;
RANGE: '..' ;
RANGE_UNTIL: '..<' ;
ELVIS: '?:' ;
LANGLE: '<' ;
RANGLE: '>' ;
REDIRECT: '|>' ;
LE: '<=' ;
GE: '>=' ;
EXCL_EQ: '!=' ;
EXCL_EQEQ: '!==' ;
AS_SAFE: 'as?' ;
EQEQ: '==' ;
EQEQEQ: '===' ;
MAP_TO: '=>' ;
SINGLE_QUOTE: '\'' ;
ELIPSIS: '...' ;

//KEYWORDS

ABSTRACT: 'abstract' ;
APPLY: 'apply';
AS: 'as' ;
BREAK: 'break' ;
CATCH: 'catch' ;
CLASS: 'class' ;
CONSTRUCTOR: 'constructor' ;
CONTINUE: 'continue' ;
DO: 'do' ;
ELSE: 'else' ;
ENUM: 'enum' ;
FINAL: 'final';
FINALLY: 'finally' ;
FOR: 'for' ;
FUN: 'fun' ; // Deprecated. Use func or meth instead.
FUNC: 'func' ;
IF: 'if' ;
IN: 'in' ;
NOT_IN: '!in' (WS | NL)+ ;
IS: 'is' ;
NOT_IS: '!is' (WS | NL)+ ;
IMPORT: 'import' ;
INCLUDE: 'include' -> pushMode(PathMode) ;
INTERFACE: 'interface' ;
INIT: 'init' ;
METH: 'meth' ;
OPEN: 'open';
OVERRIDE: 'override' ;
PACKAGE: 'package' ;
LIBRARY: 'library' -> pushMode(PathMode) ;
RETURN: 'return' ;
STATIC: 'static' ; // Deprecated. Use class val/var or func
SUPER : 'super' ;
THIS : 'this' ;
THROW: 'throw' ;
TRY: 'try' ;
VAL: 'val' ;
VAR: 'var' ;
WHILE: 'while' ;
WITH: 'with' ;

// END Keywords

// Modes

QUOTE_OPEN: '"' -> pushMode(LineString) ;
TRIPLE_QUOTE_OPEN: '"""' -> pushMode(MultiLineString) ;

COMMAND_OPEN : '$(' -> pushMode(Command) ;

// Literals

FloatLiteral: (RealLiteral | IntegerLiteral) '_'?  [fF];

DoubleLiteral: (RealLiteral | IntegerLiteral) '_'?  [dD];

LongLiteral: (DecLiteral | HexLiteral | BinLiteral) '_'?  [lL];

ByteLiteral
    : (DecLiteral | BinLiteral) '_'? [Bb]
    | HexLiteral '_' [bB] // We need the underscore to differentiate a hex digit B (or b).
    ;

ShortLiteral
    : (DecLiteral | HexLiteral | BinLiteral) '_'? [sS]
    ;

IntegerLiteral: DecLiteral;

// A real number with no suffix to indicate if it is a double or a float. Examples :
// 3.0   3_000.0  3.000_1    3_000.000_1    3.0e2  3_000e2   3e2
// FYI, It is only by convention that underscores are between triplets of digits. The grammar doesn't enfore it.
// The exponent part CANNOT contain underscores.

RealLiteral:
    ( (DecDigitNoZero DecDigit* | '0')? '.' | (DecDigitNoZero (DecDigit | '_')* DecDigit)? '.') (
        DecDigit+
        | DecDigit (DecDigit | '_')+ DecDigit
        | DecDigit+ [eE] ('+' | '-')? DecDigit+
        | DecDigit+ [eE] ('+' | '-')? DecDigit (DecDigit | '_')+ DecDigit
        | DecDigit (DecDigit | '_')+ DecDigit [eE] ('+' | '-')? DecDigit+
    )
    |
    // Now for cases where this is no decimal point, but there is an exponent.
    ( (DecDigitNoZero DecDigit* | '0') | (DecDigitNoZero (DecDigit | '_')* DecDigit)?) (
        [eE] ('+' | '-')? DecDigit+
    )
;

DecLiteral
    : ('0'
      | DecDigitNoZero DecDigit*
      | DecDigitNoZero (DecDigit | '_')+ DecDigit
      | DecDigitNoZero DecDigit* [eE] ('+' | '-')? DecDigit+
      | DecDigitNoZero (DecDigit | '_')+ DecDigit [eE] ('+' | '-')? DecDigit+
      )
    ;

fragment DecDigit
    : UNICODE_CLASS_ND
    ;

fragment DecDigitNoZero
    : UNICODE_CLASS_ND_NoZeros
    ;

fragment UNICODE_CLASS_ND_NoZeros
	: '\u0031'..'\u0039'
	| '\u0661'..'\u0669'
	| '\u06f1'..'\u06f9'
	| '\u07c1'..'\u07c9'
	| '\u0967'..'\u096f'
	| '\u09e7'..'\u09ef'
	| '\u0a67'..'\u0a6f'
	| '\u0ae7'..'\u0aef'
	| '\u0b67'..'\u0b6f'
	| '\u0be7'..'\u0bef'
	| '\u0c67'..'\u0c6f'
	| '\u0ce7'..'\u0cef'
	| '\u0d67'..'\u0d6f'
	| '\u0de7'..'\u0def'
	| '\u0e51'..'\u0e59'
	| '\u0ed1'..'\u0ed9'
	| '\u0f21'..'\u0f29'
	| '\u1041'..'\u1049'
	| '\u1091'..'\u1099'
	| '\u17e1'..'\u17e9'
	| '\u1811'..'\u1819'
	| '\u1947'..'\u194f'
	| '\u19d1'..'\u19d9'
	| '\u1a81'..'\u1a89'
	| '\u1a91'..'\u1a99'
	| '\u1b51'..'\u1b59'
	| '\u1bb1'..'\u1bb9'
	| '\u1c41'..'\u1c49'
	| '\u1c51'..'\u1c59'
	| '\ua621'..'\ua629'
	| '\ua8d1'..'\ua8d9'
	| '\ua901'..'\ua909'
	| '\ua9d1'..'\ua9d9'
	| '\ua9f1'..'\ua9f9'
	| '\uaa51'..'\uaa59'
	| '\uabf1'..'\uabf9'
	| '\uff11'..'\uff19'
	;

HexLiteral
    : '0' [xX] HexDigit (HexDigit | '_')*
    ;

fragment HexDigit
    : [0-9a-fA-F]
    ;

BinLiteral
    : '0' [bB] BinDigit (BinDigit | '_')*
    ;

fragment BinDigit
    : [01]
    ;

BooleanLiteral
    : 'true'
    | 'false'
    ;

NullLiteral
    : 'null'
    ;

Identifier
    : (Letter | '_') (Letter | '_' | DecDigit)*
    ;

// I feel like this is "ugly", but I don't know how to make it better.
// I want the "path" after the "library" keyword to contain pretty much anything except a newline and semicolon.
//Library
//    : LIBRARY Path
//    ;

// I feel like this is "ugly", but I don't know how to make it better.
// I want the "path" after the "library" keyword to contain pretty much anything except a newline and semicolon.
//Include
//    : INCLUDE Path
//    ;

// The path after a library or include keyword is anything terminated by a newline or semicolon.
//fragment Path
//    : ~('\u000A' | '\u000D' | ';')+
//    ;

LabelReference
    : '@' Identifier
    ;

LabelDefinition
    : Identifier '@'
    ;

FieldIdentifier
    : '$' Identifier
    ;

CharacterLiteral
    : '\'' (EscapeSeq | .) '\''
    ;


fragment EscapeSeq
    : UniCharacterLiteral
    | EscapedIdentifier
    ;

fragment UniCharacterLiteral
    : '\\' 'u' HexDigit HexDigit HexDigit HexDigit
    ;

fragment EscapedIdentifier
    : '\\' ('t' | 'b' | 'r' | 'n' | '\'' | '"' | '\\' | '$')
    ;

fragment Letter
    : UNICODE_CLASS_LL
    | UNICODE_CLASS_LM
    | UNICODE_CLASS_LO
    | UNICODE_CLASS_LT
    | UNICODE_CLASS_LU
    | UNICODE_CLASS_NL
    ;


mode PathMode;
Path: ~('\u000A' | '\u000D' | ';')+ ->popMode;

mode Inside ;

Inside_RPAREN: ')' -> popMode, type(RPAREN) ;
Inside_RSQUARE: ']' -> popMode, type(RSQUARE);
Inside_BARRELCLOSE: ']]' -> popMode, type(BARRELCLOSE);

Inside_LPAREN: LPAREN -> pushMode(Inside), type(LPAREN) ;
Inside_LSQUARE: LSQUARE -> pushMode(Inside), type(LSQUARE) ;
Inside_BARRELOPEN: BARRELOPEN -> pushMode(Inside), type(BARRELOPEN) ;
Inside_COMMAND_OPEN: COMMAND_OPEN -> pushMode(Command), type(COMMAND_OPEN) ;

Inside_LCURL: LCURL -> type(LCURL) ;
Inside_RCURL: RCURL -> type(RCURL) ;
Inside_DOT: DOT -> type(DOT) ;
Inside_COMMA: COMMA  -> type(COMMA) ;
Inside_POWER: POWER-> type(POWER) ;
Inside_MULT: MULT -> type(MULT) ;
Inside_REM: REM  -> type(REM) ;
Inside_FLOOR_DIV: FLOOR_DIV -> type(FLOOR_DIV) ;
Inside_DIVIDE: DIVIDE -> type(DIVIDE) ;
Inside_BACKSLASH: BACKSLASH -> type(BACKSLASH) ;
Inside_ADD: ADD  -> type(ADD) ;
Inside_SUB: SUB  -> type(SUB) ;
Inside_INCR: INCR  -> type(INCR) ;
Inside_DECR: DECR  -> type(DECR) ;
Inside_ARROW: ARROW  -> type(ARROW) ;
Inside_CONJ: CONJ  -> type(CONJ) ;
Inside_DISJ: DISJ  -> type(DISJ) ;
Inside_PIPE: PIPE  -> type(PIPE) ;
Inside_EXCL: EXCL  -> type(EXCL) ;
Inside_COLON: COLON  -> type(COLON) ;
Inside_COLON_COLON: COLON_COLON  -> type(COLON_COLON) ;
Inside_CURRIED: CURRIED  -> type(CURRIED) ;
Inside_SEMICOLON: SEMICOLON  -> type(SEMICOLON) ;
Inside_ASSIGNMENT: ASSIGNMENT  -> type(ASSIGNMENT) ;
Inside_ADD_ASSIGNMENT: ADD_ASSIGNMENT  -> type(ADD_ASSIGNMENT) ;
Inside_SUB_ASSIGNMENT: SUB_ASSIGNMENT  -> type(SUB_ASSIGNMENT) ;
Inside_MULT_ASSIGNMENT: MULT_ASSIGNMENT  -> type(MULT_ASSIGNMENT) ;
Inside_FLOOR_DIV_ASSIGNMENT: FLOOR_DIV_ASSIGNMENT  -> type(FLOOR_DIV_ASSIGNMENT) ;
Inside_DIVIDE_ASSIGNMENT: DIVIDE_ASSIGNMENT  -> type(DIVIDE_ASSIGNMENT) ;
Inside_REM_ASSIGNMENT: REM_ASSIGNMENT  -> type(REM_ASSIGNMENT) ;
Inside_RANGE: RANGE  -> type(RANGE) ;
Inside_RANGE_UNTIL: RANGE_UNTIL  -> type(RANGE_UNTIL) ;
Inside_ELVIS: ELVIS  -> type(ELVIS) ;
Inside_LANGLE: LANGLE  -> type(LANGLE) ;
Inside_RANGLE: RANGLE  -> type(RANGLE) ;
Inside_RANGLE_RANGLE: REDIRECT  -> type(REDIRECT) ;
Inside_LE: LE  -> type(LE) ;
Inside_GE: GE  -> type(GE) ;
Inside_MAP_TO: MAP_TO -> type(MAP_TO) ;
Inside_EXCL_EQ: EXCL_EQ  -> type(EXCL_EQ) ;
Inside_EXCL_EQEQ: EXCL_EQEQ  -> type(EXCL_EQEQ) ;
Inside_IS: IS -> type(IS) ;
Inside_NOT_IS: NOT_IS -> type(NOT_IS) ;
Inside_NOT_IN: NOT_IN -> type(NOT_IN) ;
Inside_AS: AS -> type(AS) ;
Inside_AS_SAFE: AS_SAFE  -> type(AS_SAFE) ;
Inside_EQEQ: EQEQ  -> type(EQEQ) ;
Inside_EQEQEQ: EQEQEQ  -> type(EQEQEQ) ;
Inside_SINGLE_QUOTE: SINGLE_QUOTE  -> type(SINGLE_QUOTE) ;
Inside_QUOTE_OPEN: QUOTE_OPEN -> pushMode(LineString), type(QUOTE_OPEN) ;
Inside_TRIPLE_QUOTE_OPEN: TRIPLE_QUOTE_OPEN -> pushMode(MultiLineString), type(TRIPLE_QUOTE_OPEN) ;

Inside_VAR: VAR -> type(VAR) ;
Inside_VAL: VAL -> type(VAL) ;
Inside_ELIPSIS: ELIPSIS -> type(ELIPSIS) ;
Inside_IN: IN -> type(IN) ;
Inside_RETURN: RETURN -> type(RETURN) ;
Inside_APPLY : APPLY -> type(APPLY) ;
Inside_CONTINUE: CONTINUE -> type(CONTINUE) ;
Inside_BREAK: BREAK -> type(BREAK) ;
Inside_IF: IF -> type(IF) ;
Inside_ELSE: ELSE -> type(ELSE) ;
Inside_TRY: TRY -> type(TRY) ;
Inside_CATCH: CATCH -> type(CATCH) ;
Inside_FINALLY: FINALLY -> type(FINALLY) ;
Inside_THROW: THROW -> type(THROW) ;
Inside_FOR: FOR -> type(FOR) ;
Inside_DO: DO -> type(DO) ;
Inside_WHILE: WHILE -> type(WHILE) ;
Inside_WITH: WITH -> type(WITH) ;

Inside_BooleanLiteral: BooleanLiteral -> type(BooleanLiteral) ;
Inside_ByteLiteral: ByteLiteral -> type(ByteLiteral) ;
Inside_CharacterLiteral: CharacterLiteral -> type(CharacterLiteral) ;
Inside_ShortLiteral: ShortLiteral -> type(ShortLiteral) ;
Inside_IntegerLiteral: IntegerLiteral -> type(IntegerLiteral) ;
Inside_LongLiteral: LongLiteral -> type(LongLiteral) ;
Inside_BinLiteral: BinLiteral -> type(BinLiteral) ;
Inside_DecLiteral: DecLiteral -> type(DecLiteral) ;
Inside_HexLiteral: HexLiteral -> type(HexLiteral) ;
Inside_FloatLiteral: FloatLiteral -> type(FloatLiteral) ;
Inside_DoubleLiteral: DoubleLiteral -> type(DoubleLiteral) ;
Inside_RealLiteral: RealLiteral -> type(RealLiteral) ;
Inside_NullLiteral: NullLiteral -> type(NullLiteral) ;


Inside_Identifier: Identifier -> type(Identifier) ;
Inside_LabelReference: LabelReference -> type(LabelReference) ;
Inside_LabelDefinition: LabelDefinition -> type(LabelDefinition) ;
Inside_Documentation: Documentation -> type(Documentation) ;
Inside_Comment: (LineComment | DelimitedComment) -> channel(HIDDEN) ;
Inside_WS: WS -> skip ;
Inside_NL: NL -> skip ;


mode LineString ;

QUOTE_CLOSE
    : '"' -> popMode
    ;

LineStrRef
    : FieldIdentifier
    ;

LineStrText
    : ~('\\' | '"' | '$')+ | '$'
    ;

LineStrEscapedChar
    : '\\' .
    | UniCharacterLiteral
    ;

LineStrExprStart
    : '${' -> pushMode(StringExpression)
    ;


mode MultiLineString ;

TRIPLE_QUOTE_CLOSE
    : MultiLineStringQuote? '"""' -> popMode
    ;

MultiLineStringQuote
    : '"'+
    ;

MultiLineStrRef
    : FieldIdentifier
    ;

MultiLineStrText
    :  ~('\\' | '"' | '$')+ | '$'
    ;

MultiLineStrEscapedChar
    : '\\' .
    ;

MultiLineStrExprStart
    : '${' -> pushMode(StringExpression)
    ;

MultiLineNL: NL -> skip ;


mode StringExpression ;

StrExpr_RCURL: RCURL -> popMode, type(RCURL) ;

StrExpr_LPAREN: LPAREN -> pushMode(Inside), type(LPAREN) ;
StrExpr_LSQUARE: LSQUARE -> pushMode(Inside), type(LSQUARE) ;

StrExpr_RPAREN: ')' -> type(RPAREN) ;
StrExpr_RSQUARE: ']' -> type(RSQUARE);
StrExpr_LCURL: LCURL -> pushMode(StringExpression), type(LCURL) ;
StrExpr_DOT: DOT -> type(DOT) ;
StrExpr_COMMA: COMMA  -> type(COMMA) ;
StrExpr_COMMAND_OPEN: COMMAND_OPEN -> type(COMMAND_OPEN) ;
StrExpr_POWER: POWER -> type(POWER) ;
StrExpr_MULT: MULT -> type(MULT) ;
StrExpr_REM: REM  -> type(REM) ;
StrExpr_FLOOR_DIV: FLOOR_DIV -> type(FLOOR_DIV) ;
StrExpr_DIVIDE: DIVIDE -> type(DIVIDE) ;
StrExpr_BLACKSLASH: BACKSLASH -> type(BACKSLASH) ;
StrExpr_ADD: ADD  -> type(ADD) ;
StrExpr_SUB: SUB  -> type(SUB) ;
StrExpr_INCR: INCR  -> type(INCR) ;
StrExpr_DECR: DECR  -> type(DECR) ;
StrExpr_ARROW: ARROW  -> type(ARROW) ;
StrExpr_CONJ: CONJ  -> type(CONJ) ;
StrExpr_DISJ: DISJ  -> type(DISJ) ;
StrExpr_PIPE: PIPE  -> type(PIPE) ;
StrExpr_EXCL: EXCL  -> type(EXCL) ;
StrExpr_COLON: COLON  -> type(COLON) ;
StrExpr_COLON_COLON: COLON_COLON  -> type(COLON_COLON) ;
StrExpr_CURRIED: CURRIED  -> type(CURRIED) ;
StrExpr_SEMICOLON: SEMICOLON  -> type(SEMICOLON) ;
StrExpr_ASSIGNMENT: ASSIGNMENT  -> type(ASSIGNMENT) ;
StrExpr_ADD_ASSIGNMENT: ADD_ASSIGNMENT  -> type(ADD_ASSIGNMENT) ;
StrExpr_SUB_ASSIGNMENT: SUB_ASSIGNMENT  -> type(SUB_ASSIGNMENT) ;
StrExpr_MULT_ASSIGNMENT: MULT_ASSIGNMENT  -> type(MULT_ASSIGNMENT) ;
StrExpr_FLOOR_DIV_ASSIGNMENT: FLOOR_DIV_ASSIGNMENT  -> type(FLOOR_DIV_ASSIGNMENT) ;
StrExpr_DIVIDE_ASSIGNMENT: DIVIDE_ASSIGNMENT  -> type(DIVIDE_ASSIGNMENT) ;
StrExpr_REM_ASSIGNMENT: REM_ASSIGNMENT  -> type(REM_ASSIGNMENT) ;
StrExpr_RANGE: RANGE  -> type(RANGE) ;
StrExpr_RANGE_UNTIL: RANGE_UNTIL  -> type(RANGE_UNTIL) ;
StrExpr_ELVIS: ELVIS  -> type(ELVIS) ;
StrExpr_LANGLE: LANGLE  -> type(LANGLE) ;
StrExpr_RANGLE: RANGLE  -> type(RANGLE) ;
StrExpr_RANGLE_RANGLE: REDIRECT  -> type(REDIRECT) ;
StrExpr_LE: LE  -> type(LE) ;
StrExpr_GE: GE  -> type(GE) ;
StrExpr_MAP_TO: MAP_TO  -> type(MAP_TO) ;
StrExpr_EXCL_EQ: EXCL_EQ  -> type(EXCL_EQ) ;
StrExpr_EXCL_EQEQ: EXCL_EQEQ  -> type(EXCL_EQEQ) ;
StrExpr_AS: AS -> type(IS) ;
StrExpr_IS: IS -> type(IN) ;
StrExpr_IN: IN ;
StrExpr_NOT_IS: NOT_IS -> type(NOT_IS) ;
StrExpr_NOT_IN: NOT_IN -> type(NOT_IN) ;
StrExpr_AS_SAFE: AS_SAFE  -> type(AS_SAFE) ;
StrExpr_EQEQ: EQEQ  -> type(EQEQ) ;
StrExpr_EQEQEQ: EQEQEQ  -> type(EQEQEQ) ;
StrExpr_SINGLE_QUOTE: SINGLE_QUOTE  -> type(SINGLE_QUOTE) ;
StrExpr_QUOTE_OPEN: QUOTE_OPEN -> pushMode(LineString), type(QUOTE_OPEN) ;
StrExpr_TRIPLE_QUOTE_OPEN: TRIPLE_QUOTE_OPEN -> pushMode(MultiLineString), type(TRIPLE_QUOTE_OPEN) ;

StrExpr_BooleanLiteral: BooleanLiteral -> type(BooleanLiteral) ;
StrExpr_ByteLiteral: ByteLiteral -> type(ByteLiteral) ;
StrExpr_CharacterLiteral: CharacterLiteral -> type(CharacterLiteral) ;
StrExpr_ShortLiteral: ShortLiteral -> type(ShortLiteral) ;
StrExpr_IntegerLiteral: IntegerLiteral -> type(IntegerLiteral) ;
StrExpr_LongLiteral: LongLiteral -> type(LongLiteral) ;
StrExpr_FloatLiteral: FloatLiteral -> type(FloatLiteral) ;
StrExpr_DoubleLiteral: DoubleLiteral -> type(DoubleLiteral) ;
StrExpr_RealLiteral: RealLiteral -> type(RealLiteral) ;
StrExpr_NullLiteral: NullLiteral -> type(NullLiteral) ;
StrExpr_BinLiteral: BinLiteral -> type(BinLiteral) ;
StrExpr_DecLiteral: DecLiteral -> type(DecLiteral) ;
StrExpr_HexLiteral: HexLiteral -> type(HexLiteral) ;

StrExpr_Identifier: Identifier -> type(Identifier) ;
StrExpr_LabelReference: LabelReference -> type(LabelReference) ;
StrExpr_LabelDefinition: LabelDefinition -> type(LabelDefinition) ;
StrExpr_Comment: (LineComment | DelimitedComment) -> channel(HIDDEN) ;
StrExpr_WS: WS -> skip ;
StrExpr_NL: NL -> skip ;

mode Command ;

COMMAND_CLOSE
    : ')' -> popMode
    ;

CommandRef
    : FieldIdentifier
    ;

CommandText
:  ~('\\' | ')' | '$')+ | '$'
    ;

CommandEscapedChar
    : '\\' .
    | UniCharacterLiteral
    ;

CommandExprStart
    : '${' -> pushMode(StringExpression)
    ;
